.pairs <- function(str) {
	s <- strsplit(str, "")[[1]]
	c1 <- c2 <- integer()
	parens <- square <- curly <- straight <- A <- B <- C <- D <- E <- integer()
	for (j in seq_along(s)) {
		if (s[j]=="(") {
			parens <- c(parens, j)
		} else if (s[j]==")") {
			if (length(parens)==0)
				stop("CONS has imbalanced pairings '(' and ')'.")
			c1 <- c(c1, parens[length(parens)])
			c2 <- c(c2, j)
			length(parens) <- length(parens) - 1L
		} else if (s[j]=="[") {
			square <- c(square, j)
		} else if (s[j]=="]") {
			if (length(square)==0)
				stop("CONS has imbalanced pairings '[' and ']'.")
			c1 <- c(c1, square[length(square)])
			c2 <- c(c2, j)
			length(square) <- length(square) - 1L
		} else if (s[j]=="{") {
			curly <- c(curly, j)
		} else if (s[j]=="}") {
			if (length(curly)==0)
				stop("CONS has imbalanced pairings '{' and '}'.")
			c1 <- c(c1, curly[length(curly)])
			c2 <- c(c2, j)
			length(curly) <- length(curly) - 1L
		} else if (s[j]=="<") {
			straight <- c(straight, j)
		} else if (s[j]==">") {
			if (length(straight)==0)
				stop("CONS has imbalanced pairings '<' and '>'.")
			c1 <- c(c1, straight[length(straight)])
			c2 <- c(c2, j)
			length(straight) <- length(straight) - 1L
		} else if (s[j]=="A") {
			A <- c(A, j)
		} else if (s[j]=="a") {
			if (length(A)==0)
				stop("CONS has imbalanced pairings 'A' and 'a'.")
			c1 <- c(c1, A[length(A)])
			c2 <- c(c2, j)
			length(A) <- length(A) - 1L
		} else if (s[j]=="B") {
			B <- c(B, j)
		} else if (s[j]=="b") {
			if (length(B)==0)
				stop("CONS has imbalanced pairings 'B' and 'b'.")
			c1 <- c(c1, B[length(B)])
			c2 <- c(c2, j)
			length(B) <- length(B) - 1L
		} else if (s[j]=="C") {
			C <- c(C, j)
		} else if (s[j]=="c") {
			if (length(C)==0)
				stop("CONS has imbalanced pairings 'C' and 'c'.")
			c1 <- c(c1, C[length(C)])
			c2 <- c(c2, j)
			length(C) <- length(C) - 1L
		} else if (s[j]=="D") {
			D <- c(D, j)
		} else if (s[j]=="d") {
			if (length(D)==0)
				stop("CONS has imbalanced pairings 'D' and 'd'.")
			c1 <- c(c1, D[length(D)])
			c2 <- c(c2, j)
			length(D) <- length(D) - 1L
		} else if (s[j]=="E") {
			E <- c(E, j)
		} else if (s[j]=="e") {
			if (length(E)==0)
				stop("CONS has imbalanced pairings 'E' and 'e'.")
			c1 <- c(c1, E[length(E)])
			c2 <- c(c2, j)
			length(E) <- length(E) - 1L
		} # else ignore higher-order pseudoknots
	}
	
	if (length(parens) != 0)
		stop("CONS has imbalanced pairings '(' and ')'.")
	if (length(square) != 0)
		stop("CONS has imbalanced pairings '[' and ']'.")
	if (length(curly) != 0)
		stop("CONS has imbalanced pairings '{' and '}'.")
	if (length(straight) != 0)
		stop("CONS has imbalanced pairings '<' and '>'.")
	if (length(A) != 0)
		stop("CONS has imbalanced pairings 'A' and 'a'.")
	if (length(B) != 0)
		stop("CONS has imbalanced pairings 'B' and 'b'.")
	if (length(C) != 0)
		stop("CONS has imbalanced pairings 'C' and 'c'.")
	if (length(D) != 0)
		stop("CONS has imbalanced pairings 'D' and 'd'.")
	if (length(E) != 0)
		stop("CONS has imbalanced pairings 'E' and 'e'.")
	
	return(cbind(c1, c2))
}

.getStruct <- function(RNA, STRUCT) {
	v <- vmatchPattern("-", RNA)
	l <- lapply(v, start)
	if (any((width(RNA) - lengths(l)) != lengths(STRUCT)))
		stop("All sequences in RNA must be the same length as their corresponding structure in STRUCT.")
	STRUCT2 <- lapply(width(RNA), rep, x=-1)
	for (k in seq_along(STRUCT2)) {
		if (length(l[[k]]) > 0) {
			map <- STRUCT2[[k]]
			map[-l[[k]]] <- seq_len(length(map) - length(l[[k]]))
			w <- which(STRUCT[[k]] > 0) # paired positions
			if (length(w) > 0)
				STRUCT[[k]][w] <- match(STRUCT[[k]][w], map)
			STRUCT2[[k]][-l[[k]]] <- STRUCT[[k]]
		} else {
			STRUCT2[[k]] <- STRUCT[[k]]
		}
	}
	STRUCT <- matrix(unlist(STRUCT2), byrow=TRUE, nrow=length(STRUCT2))
	# 0 = dot; -1 = gap; NA = mask; > 0 = paired
}

.errorCheckRNA <- function(RNA) {
	if (!is(RNA, "RNAStringSet") && !is(RNA, "DNAStringSet"))
		stop("RNA must be a DNAStringSet or RNAStringSet.")
	if (length(unique(width(RNA))) != 1)
		stop("All RNA sequences must have the same width (be aligned).")
	if (width(RNA)[1]==0)
		stop("The alignment must have a non-zero width.")
	invisible(NULL)
}

.errorCheckSTRUCT <- function(STRUCT) {
	if (!is.list(STRUCT))
		stop("STRUCT must be of type 'list'.")
	if (!all(sapply(STRUCT, is.numeric)))
		stop("All elements of STRUCT must be numeric.")
	invisible(NULL)
}

.errorCheckCONS <- function(CONS) {
	if (!is.character(CONS))
		stop("CONS must be of type 'character'.")
	if (length(CONS) > 1)
		stop("CONS must have length 1.")
	if (grepl("[^\\.\\(\\)\\[\\]\\{\\}\\<\\>AaBbCcDdEe]", CONS))
		stop("CONS contains an invalid character.")
}

ScoreAlignment <- function(RNA, STRUCT) {
	if (length(RNA) < 2)
		stop("RNA must contain at least 2 sequences.")
	.errorCheckRNA(RNA)
	.errorCheckSTRUCT(STRUCT)
	
	STRUCT <- .getStruct(RNA, STRUCT)
	
	# mask any terminal gaps
	t <- TerminalChar(RNA)
	for (k in seq_along(RNA)) {
		if (t[k, 1] > 0)
			STRUCT[k, seq_len(t[k, 1])] <- NA_integer_
		if (t[k, 2] > 0)
			STRUCT[k, ncol(STRUCT) - seq_len(t[k, 2]) + 1] <- NA_integer_
	}
	
	score <- apply(STRUCT,
		2,
		function(x) {
			n <- table(na.omit(x))
			if (length(n) > 0) {
				y <- sum(n*(n - 1)) # matches
			} else {
				y <- 0
			}
			z <- sum(n*(sum(n) - n)) # mismatches
			y/(y + z)
		})
	score <- weighted.mean(score, # fraction of diagonal in confusion matrix
		colSums(consensusMatrix(RNA)[1:15,]), # column occupancy
		na.rm=TRUE)
	
	return(score)
}

ScoreStructure <- function(RNA, STRUCT, CONS) {
	if (length(RNA) < 1)
		stop("RNA must contain at least 1 sequence.")
	.errorCheckRNA(RNA)
	.errorCheckSTRUCT(STRUCT)
	.errorCheckCONS(CONS)
	if (nchar(CONS) != width(RNA)[1])
		stop("CONS must be the same width as RNA.")
	
	STRUCT <- .getStruct(RNA, STRUCT)
	
	p <- .pairs(CONS)
	pairings <- integer(nchar(CONS))
	pairings[p[, 1]] <- p[, 2]
	pairings[p[, 2]] <- p[, 1]
	
	scores <- numeric(length(RNA))
	for (k in seq_along(scores)) {
		w <- which(STRUCT[k,] >= 0)
		scores[k] <- sum(STRUCT[k, w]==pairings[w])/length(w)
	}
	
	return(mean(scores))
}
